"""
Set of convertes that returns a Resource class. Required by Restful
implementation
"""

from __future__ import absolute_import

def __import(mod, names=None):
    """
    Names are imported from mod immediately, and then added to the
    global scope.  (It is equivalent to 'from mod import <names>')
    """
    # No lazy importing, import everything immediately.
    omod = __import__(mod, globals(), fromlist=names)
    if names:
        # from mod import <names>
        for name in names:
            globals()[name] = getattr(omod, name)
    else:
        # import mod
        globals()[mod] = omod


def __find(path, expr="*.py*"):
    """
    return a tuple of valid modules form __import__
    """
    import os
    import glob
    import itertools

    index, pattern = 0, os.path.join(path, expr)
    if not os.path.isabs(path):
        path = os.path.dirname(__file__)
        index = len(path)
        pattern = os.path.abspath(os.path.join(path, pattern))

    func = lambda x: x[:x.rfind('.')].replace(os.sep, '.')[index + 1:]
    modules = set(itertools.imap(func, glob.glob(pattern)))
    # remove '__init__' if present
    modules.remove("__init__")
    return modules

#pylint: disable-msg=W0141
map(__import, __find(".",  expr="*.py*"))



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
multipart

multipart marshaller
"""

from __future__ import absolute_import

__author__ = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.
import hashlib
import cStringIO

__all__ = []

# Project requirements

# local submodule requirements
from ..converters import ConverterMixin


class Part(object):

    EOL="\r\n".encode('latin-1')

    def __init__(self, boundary, name, contents):
        self.boundary = boundary
        self.name     = name
        self.contents = contents

    def write(self, output):
        raise NotImplementedError

        
class ParamPart(Part):
    def write(self, output):
        output.writelines(("--", self.boundary, self.EOL))
        output.writelines(('Content-Disposition: form-data; name="', self.name, '"', self.EOL))
        output.writelines((self.EOL, self.contents, self.EOL))


class FilePart(Part):
    def write(self, output, algorithm=None):
        # write header
        output.writelines(("--", self.boundary, self.EOL))
        output.writelines(('Content-Disposition: form-data; name="', self.name, '"; filename="', self.name, '"' ,self.EOL))
        output.writelines(('Content-Type: application/octet-stream', self.EOL))
        output.writelines(('Content-Transfer-Encoding: binary', self.EOL))
        output.write(self.EOL)

        # write file contents
        digest = algorithm() or None
        for chunk in iter(lambda: self.contents.read(8192), b''):
            output.write(chunk)
            digest and digest.update(chunk)

        # end header
        output.write(self.EOL)
        return digest

    
class MultiPartConverter(ConverterMixin):
    """Multipart form encoder"""

    types = ['multipart/form-data']


    def __init__(self, boundary=None):
        # Only create a boundary property if boundary is defined
        if boundary is not None:
            self.boundary = boundary[0].split('=')[1]
        ConverterMixin.__init__(self)


    def marshal(self, content):
        """Produces a well formated multipart"""

        assert isinstance(content, dict)
        assert self.boundary

        # ouput buffer
        output = cStringIO.StringIO()

        # parse params
        for key, value in content.iteritems():
            # write header
            if hasattr(value, "writelines"):
                flpart = FilePart(self.boundary, key, value)
                digest = flpart.write(output, algorithm=hashlib.sha1)
                
                # prepare to write digest param
                key   = key + "_sha1sum"
                value = digest.hexdigest()

            # it's a param or we are writing sum
            param = ParamPart(self.boundary, key, value)
            param.write(output)

        output.writelines(("--", self.boundary, "--", Part.EOL))
        retval = output.getvalue()
        
        # free buffer and return contents
        output.close()
        return retval




#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
resource

Represents a generic gateway to AMQP implementation
"""

from __future__ import absolute_import

__author__ = "caelum - http://caelum.com.br"
__modified_by__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.

__all__ = ['Resource']

# Project requirements

# local submodule requirements

class ResourceError(Exception):
    """Resource exception"""

#pylint: disable-msg=R0921
class Resource(object):
    """Prepares a connection. only get a pika connection is lazy created"""

    def links(self):
        """Returns a list of all links."""
        raise NotImplementedError

    def link(self, rel):
        """Return a Link with rel."""
        raise NotImplementedError

    def error(self):
        """
        Return error if any or None if no error is present or isn't
        suported by format
        """
        return None



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
request
"""

from __future__ import absolute_import

__author__ = "caelum - http://caelum.com.br"
__modified_by__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.

__all__ = ['Request']

# Project requirements
from .parser import Parser

class Request(object):
    """An HTTP request"""

    def __init__(self, config):
        self._config = config

    def __call__(self, callback=None, params=None, **payload):
        """
        Perform the request.The optional payload argument is sent to
        the server
        """
        env = {
            'params'  : params,
            'payload' : payload, 
            }

        procs = list(self._config.processors)
        return Parser(procs).follow(callback, self._config, env)



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
xml

XML converter and resource
"""

from __future__ import absolute_import

__author__ = "caelum - http://caelum.com.br"
__modified_by__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.
from xml.etree import ElementTree

__all__ = []

# local submodule requirements
from ..resource import Resource
from ..links import Links, Link
from ..converters import ConverterMixin


class XMLResource(Resource):
    """
    This resource is returned when a XML is unmarshalled.
    """

    def __init__ (self, element_tree):
        super(XMLResource, self).__init__()

        self.element_tree = element_tree
        self._links = Links(self._parse_links())
        self._enhance_element_tree()

    def _enhance_element_tree(self):
        """
        Enables access to XMLResources attributes with 'dot'.
        """
        setattr(self, "tag", self.element_tree.tag)

        for root_child in list(self.element_tree):
            tag = root_child.tag
            if tag != 'link':
                if len(self.element_tree.findall(root_child.tag)) > 1:
                    setattr(self, tag, self.element_tree.findall(tag))
                elif len(list(root_child)) == 0:
                    setattr(self, tag, root_child.text)
                else:
                    setattr(self, tag, self.element_tree.find(tag))

        for element in self.element_tree.getiterator():
            for child in list(element):
                if len(element.findall(child.tag)) > 1:
                    setattr(element, child.tag, element.findall(child.tag))
                elif len(list(child)) == 0:
                    setattr(element, child.tag, child.text)
                else:
                    setattr(element, child.tag, element.find(child.tag))

    def _parse_links(self):
        """
        Find links in a ElementTree
        """
        links = []
        for element in self.element_tree.getiterator('link'):
            link = Link({
                    'href': element.attrib.get('href'),
                    'rel':  element.attrib.get('rel'),
                    'type': element.attrib.get('type') or 'application/xml',
            })
            links.append(link)
        return links

    def links(self):
        return self._links

    def link(self, rel):
        return self.links().get(rel)



class XmlConverter(ConverterMixin):
    """
    Converts objects from and to XML.
    """
    
    types = ['application/xml', 'text/xml', 'xml']

    def __init__(self):
        ConverterMixin.__init__(self)

    def marshal(self, content):
        """
        Produces a XML representation of the given content.
        """
        return ElementTree.tostring(self._dict_to_etree(content))

    def _dict_to_etree(self, content):
        """
        Receives a dictionary and converts to an ElementTree
        """
        tree = ElementTree.Element(content.keys()[0])
        self._dict_to_etree_rec(content[content.keys()[0]], tree)
        return tree

    def _dict_to_etree_rec(self, content, tree):
        """
        Auxiliar function of _dict_to_etree_rec
        """
        if type(content) == dict:
            for key, value in content.items():
                element = ElementTree.Element(key)
                self._dict_to_etree_rec(value, element)
                tree.append(element)
        else:
            tree.text = str(content)

    def unmarshal(self, content):
        """
        Produces an ElementTree object for a given XML content.
        """
        element = ElementTree.fromstring(content.read())
        return XMLResource(element)




#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
BaseAPI
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.
from itertools import ifilter

__all__ = ['BaseAPI']

# Project requirements
from .restfulie import Restfulie


class BaseAPI(object):
    """
    Derive from here Custom API Implementations. All implementations
    MUST be stateless
    """

    API_BASE = None
    FLAVORS  = None
    CHAIN    = None

    #pylint: disable-msg=W0613
    @classmethod
    def _get(cls, client, auth, endpoint, args, callback):
        """Implementation of verb GET"""
        return Restfulie.at(cls.API_BASE + endpoint, cls.FLAVORS, cls.CHAIN) \
            .auth(client.credentials, method=auth)                           \
            .get(callback=callback, params=args)

    @classmethod
    def _post(cls, client, auth, endpoint, args, callback):
        """Implementation of verb POST"""

        #default to form-urlencode. If somthing that smells like a
        #file (has read function) is pased in args, encode it as
        #multipart form

        encode_type = "form"
        if any((hasattr(arg, "read") for arg in args.itervalues())):
            encode_type = "multipart"
            
        return Restfulie.at(cls.API_BASE + endpoint, cls.FLAVORS, cls.CHAIN) \
            .as_(encode_type)                                                \
            .auth(client.credentials, method=auth)                           \
            .post(callback=callback, **args)

    @classmethod
    def invoke(cls, client, call, args, callback=None):
        """Invoke method"""

        # check that requirements are passed
        path = call["endpoint"]
        requirements = call.get("required", ())
        for req in ifilter(lambda arg: arg not in args, requirements):
            err = "Missing arg '%s' for '%s'" % (req, path)
            raise AttributeError(err)

        # build endpoint
        endpoint = path % args
        auth     = call.get("auth")
        verb     = getattr(cls, "_" + call["method"])

        # remove used args
        if args:
            func = lambda x: '%%(%s)' % x[0] not in path
            args = dict(ifilter(func, args.iteritems()))
                
        # invoke
        return verb(client, auth, endpoint, args, callback)

    @classmethod
    def query(cls, client, call, args, callback=None):
        """Implements RECOVER on a CRUD model"""
        return cls.invoke(client, call, args, callback)

        




#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
links

Hipermedia objects
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.
import hal

__all__ = ['Link', 'Links']

# Project requirements

# local submodule requirements

class LinkError(Exception):
    """Define a missing link error"""


#pylint: disable-msg=R0903
class Link(hal.Link):
    """Link represents generic link. You can follow it"""

    def follow(self):
        """Return a DSL object with the Content-Type set"""
        from .configuration import Configuration
        return Configuration(self.href).as_(self.content_type)


#pylint: disable-msg=R0903
class Links(hal.Links):
    """Links a simple Link """



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
json

Json converter and resource
"""

from __future__ import absolute_import

__author__ = "caelum - http://caelum.com.br"
__modified_by__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.
from itertools import ifilter
import simplejson as json

__all__ = []

# Project requirements

# local submodule requirements
from ..resource import Resource
from ..links import Links, Link
from ..converters import ConverterMixin


class JsonResource(Resource):
    """This resource is returned when a JSON is unmarshalled"""

    def __init__(self, data):
        """JsonResource attributes can be accessed with 'dot'"""
        super(JsonResource, self).__init__()
        assert isinstance(data, dict)
        for key, value in ifilter(lambda x: x[0] != 'link', data.iteritems()):
            if isinstance(value, (list, tuple,)):
                setattr(self, key, [self.parse(element) for element in value])
            else:
                setattr(self, key, self.parse(value))

        # store dict
        self._dict  = data
        self._links = None

    def __len__(self):
        return len(self._dict)

    def _find_dicts_in_dict(self, data):
        """Get all dictionaries on a structure and returns a list of it"""
        dicts = []
        if isinstance(data, dict):
            dicts.append(data)
            for value in data.itervalues():
                dicts.extend(self._find_dicts_in_dict(value))
        return dicts

    def _parse_links(self, data):
        """Find links on JSON dictionary"""
        retval, dct_filter = [], lambda x: '_links' in x
        for dct in ifilter(dct_filter, self._find_dicts_in_dict(data)):
            #Set a json as the default content-type for this link if
            #no one has been set by the server
            #pylint:disable-msg=W0106
            for key, link in dct['_links'].iteritems():
                link = dict(link)
                link.setdefault('rel', key)
                link.setdefault('type', 'application/json')
                retval.append(Link(**link))
        return retval

    def link(self, rel):
        return self.links().get(rel)

    def links(self):
        if not self._links:
            self._links = Links(self._parse_links(self._dict))
        return self._links

    def error(self):
        if not self._error:
            self._error = self._dict.get('_error')
        return self._error
        
    @classmethod
    def parse(cls, value):
        """Pyhtonize a Json value"""
        return cls(value) if isinstance(value, dict) else value

class JsonConverter(ConverterMixin):
    """Converts objects from and to JSON"""

    types = ['application/json', 'text/json', 'json']

    def __init__(self):
        ConverterMixin.__init__(self)

    #pylint: disable-msg=R0201
    def marshal(self, content):
        """Produces a JSON representation of the given content"""
        return json.dumps(content)

    #pylint: disable-msg=R0201
    def unmarshal(self, json_content):
        """Produces an object for a given JSON content"""
        return JsonResource(json.loads(json_content.read()))





#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
processors
"""

from __future__ import absolute_import

__author__ = "caelum - http://caelum.com.br"
__modified_by__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.
from base64 import b64encode
from urllib import splittype, splithost

from tornado.gen import Task, engine
from tornado.httputil import url_concat
from tornado.httpclient import AsyncHTTPClient, HTTPClient, HTTPError

__all__ = []

# Project requirements
from oauth2 import Request, Consumer, Token, SignatureMethod_HMAC_SHA1

try:
    from urlparse import parse_qs
except ImportError:
    # fall back for Python 2.5
    from cgi import parse_qs


# local submodule requirements
from .processor import AuthMixin

class AuthError(Exception):
    """Auth exception"""

class HandShakeError(Exception):
    """Error on auth process"""
    def __init__(self, response):
        Exception.__init__(self, response.error)
        self.response = response
    
#pylint: disable-msg=R0903
class BasicAuth(AuthMixin):
    """Processor responsible for making HTTP simple auth"""

    implements = "plain"

    def authorize(self, credentials, request, env, callback):
        self.authorize_sync(credentials, request, env)
        callable(callback) and callback()

    def authorize_sync(self, credentials, request, env):
        creden = credentials.to_list("consumer_key", "consumer_secret")
        encode = b64encode("%s:%s" % creden)
        request.headers['authorization'] = 'Basic %s' % encode


class OAuthMixin(AuthMixin):
    """ oauth method """

    @property
    def request_url(self):
        """Get request_token url according to OAuth 1.0 specs"""
        raise NotImplementedError

    @property
    def access_url(self):
        """Get access_token url according to OAuth 1.0 specs"""
        raise NotImplementedError

    @property
    def authorize_url(self):
        """Get authorize url according to OAuth 1.0 specs"""
        raise NotImplementedError

    @property
    #pylint: disable-msg=W0201
    def method(self):
        """Get method used to sign oauth requests"""
        if not hasattr(self, "_method"):
            self._method = SignatureMethod_HMAC_SHA1()
        return self._method

    ##
    # Fetch methods (HTTP/HTTPS)
    #
    def _fetch(self, consumer, token, params, uri, callback):
        """Send request async"""
        def on_response(response):
            if not response.error:
                return callback(Token.from_string(response.buffer.read()))
            raise HandShakeError(response)

        # process
        request = self._get_request(consumer, token, params, uri)
        AsyncHTTPClient().fetch(request.to_url(),
            callback=on_response,
            method=request.method,
            headers=request.to_header())

    def _fetch_sync(self, consumer, token, params, uri):
        """Send request sync"""
        request = self._get_request(consumer, token, params, uri)
        response = HTTPClient().fetch(uri,
            method=request.method,
            headers=request.to_header())
        return Token.from_string(response.buffer.read())

    ##
    # OAuth related (sugar) protected methods
    #
    @staticmethod
    def _get_realm(uri):
        """ calculate realm """
        schema, rest = splittype(uri)
        hierpart = ''
        if rest.startswith('//'):
            hierpart = '//'
        host, rest = splithost(rest)
        return schema + ':' + hierpart + host

    def _get_request(self, consumer, token, params, uri):
        """Prepare an oauth request based on arguments"""
        request = Request.from_consumer_and_token(
            consumer, token, parameters=params, http_url=uri)
        request.sign_request(self.method, consumer, token)
        return request

    def _get_consumer(self, credentials):
        """Prepare and store consumer based on oauth arguments"""
        if 'consumer' not in credentials.store(self.implements):
            creds = credentials.to_list("consumer_key", "consumer_secret")
            assert all(creds)
            credentials.store(self.implements)['consumer'] = Consumer(*creds)
        return credentials.store(self.implements)['consumer']

    def _get_token(self, credentials):
        """Prepare and store a token based on credentials"""
        if 'token' not in credentials.store(self.implements):
            creds = credentials.to_list("token_key", "token_secret")
            if all(creds):
                credentials.store(self.implements)['token'] = Token(*creds)
        return credentials.store(self.implements).get('token', None)

    ##
    # OAuth Public token adquisition methods
    #
    def request_token(self, credentials, callback=None):
        """Implements first stage on OAuth  dance async"""
        self._fetch(                                     \
            self._get_consumer(credentials),             \
            None, credentials.to_dict("oauth_callback"), \
            self.request_url, callback)

    def request_token_sync(self, credentials):
        """Implements first stage on OAuth dance sync"""
        return self._fetch_sync(                         \
            self._get_consumer(credentials),             \
            None, credentials.to_dict("oauth_callback"), \
                self.request_url)
    ###
    # OAuth authorization methods
    #
    def authorization_redirect(self, token):
        """Get the authorization URL to redirect the user"""
        request = Request.from_token_and_callback(       \
            token=token, http_url=self.authorize_url)
        return request.to_url()

    def authorization_redirect_url(self, token, callback):
        """Handle url redirection after user authenticates"""
        raise NotImplementedError

    def authorization_redirect_url_sync(self, token):
        """
        Handle url redirection after user authenticates syncronously
        """
        raise NotImplementedError

    ##
    # OAuth access methods
    #
    def access_token(self, credentials, token, verifier, callback):
        """
        After user has authorized the request token, get access token
        with user supplied verifier
        """
        self._fetch(                                     \
            self._get_consumer(credentials), token,      \
            {'oauth_verifier' : verifier}, self.access_url, callback)

    def access_token_sync(self, credentials, token, verifier):
        """
        After user has authorized the request token, get access token
        with user supplied verifier
        """
        return self._fetch_sync(                         \
            self._get_consumer(credentials), token,      \
            {'oauth_verifier' : verifier}, self.access_url)

    ###
    # XAuth stuff (2LO)
    #
    def xauth_access_token(self, credentials, callback):
        """
        Get an access token from an username and password combination.
        """
        parameters = {
            'x_auth_mode':     'client_auth',
            'x_auth_username': credentials.username,
            'x_auth_password': credentials.password,
            }

        self._fetch(                                    \
            self._get_consumer(credentials), None,      \
            parameters, self.access_url, callback)

    def xauth_access_token_sync(self, credentials):
        """
        Get an access token from an username and password combination.
        """
        parameters = {
            'x_auth_mode':     'client_auth',
            'x_auth_username': credentials.username,
            'x_auth_password': credentials.password,
            }

        self._fetch_sync(                              \
            self._get_consumer(credentials), None,     \
            parameters, self.access_url)

    ###
    # OAuth sign
    # 
    def sign(self, credentials, request, env):
        """Sign request"""
        
        #pylint: disable-msg=C0103
        POST_CONTENT_TYPE = 'application/x-www-form-urlencoded'

        consumer = self._get_consumer(credentials)
        token    = self._get_token(credentials)
        
        if not consumer or not token:
            raise AuthError("Missing oauth tokens")

        # POST
        headers  = request.headers
        if request.verb == "POST":
            assert 'content-type' in headers
            
        # Only hash body and generate oauth_hash for body if
        # Content-Type != form-urlencoded
        isform = headers.get('content-type') == POST_CONTENT_TYPE

        # process post contents if required
        body, parameters = env.get('body', ''), None
        if isform and body:
            parameters = parse_qs(body)

        # update request uri
        oauth_request = Request.from_consumer_and_token(
            consumer, token, request.verb,
            url_concat(request.uri, env["params"]),
            parameters, body, isform)

        # sign
        oauth_request.sign_request(self.method, consumer, token)

        # process body if form or uri if a get/head
        if isform:
            env['body'] = oauth_request.to_postdata()
        elif request.verb in ('GET', 'HEAD',):
            # remove params and update uri store params
            request.uri   = oauth_request.to_url()
            env["params"] = None
        else:
            headers.update(oauth_request.to_header(               \
                    realm=self._get_realm(request.uri)))

    @engine
    def _authenticate(self, credentials, callback):
        token = None
        try:
            if credentials.callback:
                # Use PIN based OAuth
                token = yield Task(self.request_token, credentials)
                reurl = self.authorization_redirect(token)
                verfy = credentials.callback(reurl)
                token = yield Task(self.access_token, credentials, token, verfy)
            else:
                # Use XAuth
                token = yield Task(self.xauth_access_token, credentials)
            # retval
            callback(token)
        # A handshake error is also notified.
        except HandShakeError, err:
            callback(err.response.error)
        except HTTPError, err:
            callback(err)

    def _authenticate_sync(self, credentials):
        token = None
        if credentials.callback:
            # Use PIN based OAuth
            token = self.request_token_sync(credentials)
            reurl = self.authorization_redirect(token)
            verfy = credentials.callback(reurl)
            token = self.access_token_sync(credentials, token, verfy)
        else:
            # Use XAuth
            token = self.xauth_access_token_sync(credentials)
        return token

    def _update_credentials(self, credentials, token):
        # store token
        credentials.store(self.implements)['token'] = None
        if isinstance(token, Token):
            credentials.store(self.implements)['token']=token


    ##
    # Auth Main methods
    @engine
    def authorize(self, credentials, request, env, callback):
        response = None
        try:
            self.sign(credentials, request, env)
        except AuthError:
            response = yield Task(self._authenticate, credentials)
            self._update_credentials(credentials, response)
        # fetch token or response from server
        response = response or credentials.store(self.implements)['token']
        callback(response)
        
    def authorize_sync(self, credentials, request, env):
        try:
            self.sign(credentials, request, env)
        except AuthError:
            # fetch token
            token = self._authenticate_sync(credentials)
            self._update_credentials(credentials, token)




#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
multipart

multipart marshaller
"""

from __future__ import absolute_import

__author__ = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.
import urllib

__all__ = []

# Project requirements

# local submodule requirements
from ..converters import ConverterMixin


class MultiPartConverter(ConverterMixin):
    """Multipart form encoder"""

    types = ['application/x-www-form-urlencoded']

    def __init__(self):
        ConverterMixin.__init__(self)

    def marshal(self, content):
        """Produces a well formaed form"""
        return urllib.urlencode(content)



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
converter
"""

from __future__ import absolute_import

__author__ = "caelum - http://caelum.com.br"
__modified_by__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.

__all__ = ['Converters', 'ConverterMixin']

# Project requirements

# local submodule requirements

class ConverterError(Exception):
    """Resource exception"""

class Converters(object):
    """Utility methods for converters."""

    types = {}

    @staticmethod
    def register(a_type, converter):
        """Register a converter for the given type"""
        Converters.types[a_type] = converter

    @staticmethod
    def marshaller_for(a_type):
        """Return a converter for the given type"""
        if type(a_type) in (str, unicode,):
            # common case. Throw a key error exception if no valid one
            # has been registered"
            if ";" not in a_type:
                return Converters.types[a_type]
            # Passed a composed string (';' separated)
            a_type, key = a_type.split(";"), a_type
        else:
            # Passed a list
            assert len(a_type) > 0
            a_type, key = a_type, ";".join(a_type)

        # Dinamically, create a valid converter if required for this
        # kind of element.Converters are stateless,
        return Converters.types.setdefault(   \
            key,                              \
            Converters.types[a_type[0]].__class__(a_type[1:]))


class MetaConverter(type):
    """Converter Metaclass"""

    def __init__(mcs, name, bases, dct):
        type.__init__(mcs, name, bases, dct)
        if name.endswith("Converter"):
            for a_type in mcs.types:
                Converters.register(a_type, mcs())


class ConverterMixin(object):
    """
    Abstract class to define converter classes. This class has support
    to create chained converters
    """

    __metaclass__ = MetaConverter

    def __init__(self, a_type_list=None):
        # Store next converter in chain
        self._chain = None
        assert not a_type_list or hasattr('__iter__', a_type_list)
        # Allow this class to also be used like a dead end
        if a_type_list and len(a_type_list) > 1:
            self._chain = Converters.marshaller_for(list(a_type_list)[1:])

    def marshal(self, content):
        """Does nothing"""
        return content if not self._chain else self.chain.marshal(content)

    def unmarshal(self, content):
        """Returns content without modification"""
        return content if not self._chain else self.chain.marshal(content)



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
plain

Plain converter and resource
"""

from __future__ import absolute_import

__author__ = "caelum - http://caelum.com.br"
__modified_by__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.

__all__ = []

# Project requirements

# local submodule requirements
from ..resource import Resource
from ..links import Links
from ..converters import ConverterMixin


class PlainResource(Resource):
    """This resource is returned when a PLAIN is unmarshalled"""

    def __init__(self, content):
        """PlainResource attributes can be accessed with 'dot'"""
        super(PlainResource, self).__init__()
        self._content = content
        self._links = Links([])

    def __len__(self):
        return 0

    def link(self, rel):
        return None

    def links(self):
        return self._links

    @property
    def content(self):
        """Returns simple content"""
        return self._content


class PlainConverter(ConverterMixin):
    """Dummy converter to plain text"""

    types = ['text/plain', 'text/html',]

    def __init__(self):
        ConverterMixin.__init__(self)

    def unmarshal(self, content):
        """Returns content without modification"""
        return PlainResource(content)



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
request
"""

from __future__ import absolute_import

__author__ = "caelum - http://caelum.com.br"
__modified_by__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.

__all__ = ['Response']

# Project requirements
from .converters import Converters
from .links import Links


class Response(object):
    """Handle and parse a HTTP response"""

    def __init__(self, response):
        self._response = response
        self._resource = None

    @property
    def headers(self):
        """Returns HTTP headers"""
        return self._response.headers

    @property
    def code(self):
        """Returns response code"""
        return self._response.code

    @property
    def body(self):
        """Returns a formatted body"""
        return self._response.body

    @property
    def resource(self):
        """Unmarshalled object of the response body"""
        if not self._resource:
            content_type = self._response.headers.get_list('content-type')[0]
            converter = Converters.marshaller_for(content_type.split(';')[0])
            self._resource = converter.unmarshal(self._response.buffer)
        return self._resource
    
    @property
    def links(self):
        """Returns the Links of the header"""
        if not hasattr(self, '_links'):
            self._links = self.resource.links()
            values = self._response.headers.get('link')
            self._links.update([link for link in Links.parse(values)])
        return self._links

    def link(self, rel):
        """Get a link with 'rel' from header"""
        return self.links.get(rel)

    @property
    def error(self):
        if not hasattr(self, '_error'):
            self._error = self.resource.error()
        return self._error
        



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
parser
"""

from __future__ import absolute_import

__author__ = "caelum - http://caelum.com.br"
__modified_by__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.

__all__ = ['Parser']


#pylint: disable-msg=R0903
class Parser(object):
    """Executes processors ordered by the list"""

    def __init__(self, processors):
        self._processors = processors

    def __len__(self):
        return len(self._processors)
        
    def follow(self, callback, request, env):
        """Follow chain"""
        processor = self._processors.pop(0)
        return processor.execute(callback, self, request, env)



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
credentials
"""

from __future__ import absolute_import

__author__   = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.
from itertools import ifilter

__all__ = ['Credentials']


class Credentials(object):
    """
    Base class to define required credentials to use a remote
    service
    """
    _mechanisms = {}

    def __init__(self):
        self._api_key         = None
        self._consumer_key    = None
        self._consumer_secret = None
        self._token_key       = None
        self._token_secret    = None
        self._callback        = None

    def store(self, mechanism):
        """A store for auth mechanism data"""
        return self._mechanisms.setdefault(mechanism, {})

    def to_list(self, *args):
        """Get a list of properties collected in args"""
        return [getattr(self, arg) for arg in args]

    def to_dict(self, *args):
        """Get a dict of properties collected from args"""
        dct = {}
        for arg in ifilter(lambda x: hasattr(self, x), args):
            dct[arg] = getattr(self, arg)
        return dct

    @property
    def api_key(self):
        """Get API key required to use remote service"""
        return self._api_key or ""

    @api_key.setter
    def api_key(self, value):
        """Set API key requires to use remote service"""
        self._api_key = value

    @property
    def username(self):
        """Get username"""
        return self._token_key

    @username.setter
    def username(self, value):
        """set username"""
        self._token_key = value

    @property
    def password(self):
        """Get password"""
        return self._token_secret

    @password.setter
    def password(self, value):
        """Set password"""
        self._token_secret = value

    @property
    def consumer_key(self):
        """Get required key to authenticate (login)"""
        return self._consumer_key or ""

    @consumer_key.setter
    def consumer_key(self, value):
        """Set auth key"""
        self._consumer_key = value

    @property
    def consumer_secret(self):
        """Get password"""
        return self._consumer_secret or ""

    @consumer_secret.setter
    def consumer_secret(self, value):
        """Set password"""
        self._consumer_secret = value

    @property
    def token_key(self):
        """Get token on 2LO mechanisms"""
        return self._token_key

    @token_key.setter
    def token_key(self, value):
        """Set auth key"""
        self._token_key = value

    @property
    def token_secret(self):
        """Get tolen secret"""
        return self._token_secret

    @token_secret.setter
    def token_secret(self, value):
        """Set password"""
        self._token_secret = value

    @property
    def callback(self):
        """
        Callback to be invoked when user interaction is needed on
        handshake
        """
        return self._callback

    @callback.setter
    def callback(self, value):
        """Set callback"""
        self._callback = value




#!/usr/bin/python
#-*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(name="sleipnir-restfulie",
      version="0.12",
      description="Writing hypermedia aware resource based clients and servers",
      author="Caelum Objects <http://restfulie.caelum.com.br/>, Carlos Martín",
      author_email="inean.es@gmail.com",
      url="http://restfulie.caelumobjects.com/",
      download_url="https://github.com/inean/restfulie-py",
      packages=find_packages(),      
      license="Apache 2.0",
      keywords="rest, async, tornado, http, hypermedia",
      zip_safe=True,
      install_requires= [
          "oauth2   >= 1.5",
          "tornado  >= 2.3.0",
          "hal-json >= 0.1",
      ],
      long_description="""
      CRUD through HTTP is a good step forward to using resources
      and becoming RESTful, another step further is to make use of
      hypermedia aware resources and Restfulie allows you to do it in
      Python.
      """,
      classifiers=[
          "Development Status :: 4 - Beta",
          "Environment :: Web Environment",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: MacOS :: MacOS X",
          "Operating System :: Microsoft :: Windows",
          "Operating System :: POSIX",
          "Programming Language :: Python",
      ],
  )





#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
processors
"""

from __future__ import absolute_import

__author__ = "caelum - http://caelum.com.br"
__modified_by__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.
from copy import copy
from urlparse import urlparse

__all__ = [
    'AuthenticationProcessor',
    'ExecuteRequestProcessor',
    'PayloadMarshallingProcessor',
    'RedirectProcessor',
    ]

# Project requirements
from tornado.gen import engine, Task
from tornado.httputil import url_concat
from tornado.httpclient import HTTPClient, AsyncHTTPClient

# local submodule requirements
from .converters import Converters
from .response import Response
from .request import Request


#pylint: disable-msg=R0903
class RequestProcessor(object):
    """Base class for all processors"""

    def execute(self, callback, chain, request, env):
        """Command called to be runned"""
        raise NotImplementedError('Subclasses must implement this method')


#pylint: disable-msg=R0903, R0922
class AuthenticationProcessor(RequestProcessor):
    """Abstract class for authentication methods"""

    backends = {}

    def execute(self, callback, chain, request, env):
        if len(request.credentials):
            #pylint: disable-msg=E1101
            path = urlparse(request.uri).path or "/"
            for math, method, credentials in request.credentials.itervalues():
                if not math.match(path):
                    continue
                if method not in self.backends:
                    error = "Unsupported auth mechanism '%s'" % method
                    raise NotImplementedError(error)
                return method, credentials


#pylint: disable-msg=R0903, R0922
class AuthenticationSyncProcessor(AuthenticationProcessor):
    """Abstract class for authentication methods. Used to sign"""

    def execute(self, callback, chain, request, env):
        method = AuthenticationProcessor.execute
        method, cred = method(self, callback, chain, request, env)
        method and self.backends[method].authorize_sync(cred, request, env)
        # follow if no method used
        return chain.follow(callback, request, env)


class AuthenticationAsyncProcessor(AuthenticationProcessor):
    """Abstract class for authentication methods. Async"""

    @engine
    def execute(self, callback, chain, request, env):
        assert not chain or len(chain) == 0
        method = AuthenticationProcessor.execute
        method, cred = method(self, callback, chain, request, env)
        self.backends[method].authorize(cred, request, env, callback)
        
class MetaAuth(type):
    """Auth Metaclass"""

    def __init__(mcs, name, bases, dct):
        type.__init__(mcs, name, bases, dct)
        if name.endswith("Auth"):
            implements = mcs.implements
            AuthenticationProcessor.backends[implements] = mcs()


#pylint: disable-msg=R0921
class AuthMixin(object):
    """Base class to derive authentication mechanisms"""

    __metaclass__ = MetaAuth

    def authorize(self, credentials, request, env):
        """Command called to be runned"""
        raise NotImplementedError('Subclasses must implement this method')


class ExecuteRequestProcessor(RequestProcessor):
    """
    Processor responsible for getting the body from environment and
    making a request with it.
    """

    @staticmethod
    def _sync(request, env):
        """Run blocked"""
        request  = HTTPClient()
        response = request.fetch(
            url_concat(request.uri, env["params"]),
            method=request.verb,
            body=env.get("body"), 
            headers=request.headers)
        return Response(response)

    @staticmethod
    def _async(callback, request, env):
        """Run async"""
        request  = AsyncHTTPClient()
        response = request.fetch(
            url_concat(request.uri, env["params"]),
            lambda x: callback(Response(x)),
            method=request.verb,
            body=env.get("body"), 
            headers=request.headers)
        return request

    def execute(self, callback, chain, request, env):
        return self._sync(request, env) \
            if not callable(callback)   \
            else self._async(callback, request, env)


class PayloadMarshallingProcessor(RequestProcessor):
    """Responsible for marshalling the payload in environment"""

    def execute(self, callback, chain, request, env):
        if env["payload"]:
            content_type = request.headers.get("content-type")
            marshaller   = Converters.marshaller_for(content_type)
            env["body"]  = marshaller.marshal(env["payload"])
            del(env["payload"])
        return chain.follow(callback, request, env)


class RedirectProcessor(RequestProcessor):
    """
    A processor responsible for redirecting a client to another URI
    when the server returns the location header and a response code
    related to redirecting.
    """
    REDIRECT_CODES = ['201', '301', '302']

    def _redirect(self, result):
        """Get redirection url, if any"""
        return result.headers.get("location") \
            if (result.code in self.REDIRECT_CODES) else None

    def execute(self, callback, chain, request, env):
        def _on_resource(resource):
            """resource callback"""
            assert callable(callback)
            # copy configuration and update url
            url = self._redirect(resource)
            if not url:
                return callback(resource)
            # make a new request
            config = copy(request.config)
            config.uri = url
            return Request(self)(callback)
        # chain
        return chain.follow(_on_resource, request, env)    \
            if callable(callback) else chain.follow(callback, request, env)


# Main tornado chain (Use follow_redirects so RedirectProcessor is not
# required anymore

#pylint:disable-msg=C0103
tornado_chain = [
    PayloadMarshallingProcessor(),
    AuthenticationSyncProcessor(),
#   RedirectProcessor(),
    ExecuteRequestProcessor(),
    ]

auth_chain = [
    PayloadMarshallingProcessor(),
    AuthenticationAsyncProcessor(),
]



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
configuration
"""

from __future__ import absolute_import

__author__ = "caelum - http://caelum.com.br"
__modified_by__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.
import re

__all__ = ['Configuration']

# Project requirements
from tornado.httputil import HTTPHeaders

# local submodule requirements
from .processor import tornado_chain
from .request import Request


class Configuration(object):
    """Configuration object for requests at a given URI"""

    HTTP_VERBS = [
        'delete',
        'get',
        'head',
        'options',
        'patch',
        'post',
        'put',
        'trace'
        ]

    FLAVORS = {
        'json': {
            'content-type': 'application/json',
            'accept':       'application/json',
        },
        'xml': {
            'content-type': 'application/xml',
            'accept':       'application/xml',
        },
        'plain': {
            'content-type': 'text/plain',
            'accept':       'text/plain',
        },

        # POST only flavors
        'form': {
            'content-type': 'application/x-www-form-urlencoded',
        },
        'multipart': {
            'content-type': 'multipart/form-data; boundary=AaB03x',
        },
    }

    def __init__(self, uri, flavors=None, chain=None):
        """Initialize the configuration for requests at the given URI"""
        self.uri         = uri
        self.headers     = HTTPHeaders()
        self.flavors     = flavors or ['json', 'xml']
        self.processors  = chain or tornado_chain
        self.credentials = {}
        self.verb        = None

    def __getattr__(self, value):
        """
        Perform an HTTP request. This method supports calls to the following
        methods: delete, get, head, options, patch, post, put, trace

        Once the HTTP call is performed, a response is returned (unless the
        async method is used).
        """
        if (value not in self.HTTP_VERBS):
            raise AttributeError(value)

        # store current verb to be passed to Request
        self.verb = value.upper()

        # set accept if it wasn't set previously
        if 'accept' not in self.headers:
            for flavor in self.flavors:
                if 'accept' in self.FLAVORS[flavor]:
                    self.headers.add('accept', self.FLAVORS[flavor]['accept'])

        # set form type and default if noone is present
        if 'content-type' not in self.headers and self.verb in('POST', 'PUT'):
            self.headers['content-type'] = self.FLAVORS['form']['content-type']
            
        return Request(self)

    def use(self, feature):
        """Register a feature (processor) at this configuration"""
        self.processors.insert(0, feature)
        return self

    def as_(self, flavor):
        """Set up the Content-Type"""
        if flavor in self.FLAVORS:
            self.headers.update(self.FLAVORS[flavor])
        else:
            self.headers["accept"] = flavor
            self.headers["content-type"] = flavor
        return self

    def accepts(self, flavor):
        """Configure the accepted response format"""
        if flavor in self.FLAVORS:
            flavor = self.FLAVORS[flavor]['accept']
        self.headers.add('accept', flavor)
        return self

    def auth(self, credentials, path="*", method='plain'):
        """Authentication feature. It does simple HTTP auth"""
        # already defined ?
        if path in self.credentials or method is None:
            return self
        # process a regex valid for path
        expr = "%s.*" if path.endswith('*') else "%s$"
        rmatch = re.compile(expr % path.rsplit('*', 1)[0])
        # now store it
        self.credentials[path]=(rmatch, method, credentials,)
        return self



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
client
"""

from __future__ import absolute_import

__author__   = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.
from contextlib import contextmanager
from functools import wraps

__all__ = ['Client']

# local submodule requirements
from .credentials import Credentials

class Client(object):
    """Base class to implement a remote API"""

    def __init__(self):
        self._config = Credentials()

    @contextmanager
    def configure(self):
        """Get an accessor to credentials element"""
        yield self._config

    @property
    def credentials(self):
        """Get credentials element"""
        return self._config

    @property
    def write_enabled(self):
        """Check if config has valid credentials. Return false
        otherwise"""
        raise NotImplementedError


class Extend(type):
    """
    Allow Ad-hoc inheritance. See
    http://code.activestate.com/recipes/412717-extending-classes/ for
    details
    """

    #pylint: disable-msg=W0613
    def __new__(mcs, name, bases, dct):
        prev = globals()[name]
        del dct['__module__']
        del dct['__metaclass__']
        for key, value in dct.iteritems():
            if key == "__doc__":
                continue
            setattr(prev, key, value)
        return prev


def validate(table):
    """Check that required attributes are present on method invoke"""
    #pylint: disable-msg=C0111
    def wrap_f(func):
        def new_func(instance, action, args, callback=None):
            if action not in getattr(instance, table):
                raise AttributeError("Missing action '%s'" % action)
            return func(instance, action, args, callback)
        return wraps(func)(new_func)
    return wrap_f





#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Request
"""

from __future__ import absolute_import

__author__ = "caelum - http://caelum.com.br"
__modified_by__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE.restfulie for details"

# Import here any required modules.

__all__ = ['Response']

# Project requirements
from .configuration import Configuration

#pylint: disable-msg=R0903
class Restfulie(object):
    """Restfulie entry point"""

    # pylint: disable-msg=C0103
    @staticmethod
    def at(uri, flavors=None, chain=None):
        """Create a new entry point for executing requests"""
        return Configuration(uri, flavors, chain)
